from __future__ import annotations
from coeden.specialClasses.base_node import Node



class ValueNode(Node):
    '''Repesents a node in a tree data structure, '''
    def __init__(self,key, value = None, parent = None):
        super().__init__(key,parent)
        self.value = value

    def __set_child(self, child: Node):
        if self._children is None:
            self._children = {}
        if child.key in self._children:
            raise ValueError(f'This node already has a child with key: "{child.key}".')

        child._parent = self
        self._children[child.key] = child

    def new_leaf(self, key, value = None):
        self.__set_child(ValueNode(key, value))

    def try_new_leaf(self,key, value=None):
        '''Create a node as a child of the caller node. If succesful returns true,
        if node already exist return false'''
        try:
            self.__set_child(ValueNode(key,value))
            return True
        except ValueError:
            return False

    def try_remove_leaf(self, key):
        '''Try to remove a leaf from the tree by key, if success returns True'''
        #not a leaf if it has children
        if self[key]._children is True:
            return False
        try:
            self[key].parent = None
            del self._children[key]
            return True
        except:
            return False

    def print_tree(self):
        '''Print to console a graphical representation of the tree starting in the current node'''
        iterator = self.deep_first_iter()
        node, depth = next(iterator)
        line = node.key
        if node.value is not None:
                line += f' -> {node.value}'
        print(line)
        islast = [0]
        for (node, depth) in iterator:
            if len(islast) < depth + 1:
                islast.append(0)
            if islast[depth] == 0:
                islast[depth] = len(node.parent._children)
            islast[depth] -= 1

            tree_lines = ""
            for i in range(0,depth):
                if islast[i] >= 1:
                    tree_lines += "|  "
                else:
                    tree_lines += "   "

            line = f'{tree_lines}|-- {node.key}'
            if node.value is not None:
                line += f' -> {node.value}'
            print(line)



                



import unittest
from coeden import ValueNode
from coeden.specialClasses.base_node import NoNode

class NoNodes(unittest.TestCase):
    def test_get_unexistent_node_returns_nonode(self):
        tree = ValueNode("root")
        is_nonode = isinstance(tree["ghost node"],NoNode)
        self.assertEqual(is_nonode, True)
    
    def test_nonode_is_None(self):
        tree = ValueNode("root")
        self.assertEqual(tree["unexistent node"], None)
        self.assertEqual(tree["unexistent node"].key, None)
        self.assertEqual(tree["unexistent node"].value, None)

    def test_chained_nonodes(self):
        tree = ValueNode("root")
        nonode = tree["unexisting node"]["ghost node"]["magic node"]
        is_nonode = isinstance(nonode,NoNode)
        self.assertEqual(is_nonode, True)

    def test_add_nonode(self):
        tree = ValueNode("root")
        tree["ethereal node"].create_all()
        self.assertEqual(tree["ethereal node"].key, "ethereal node")

    def test_add_multiple_nonodes(self):
        tree = ValueNode("root")
        tree["new 1"]["new 2"].create_all()

        self.assertEqual(tree["new 1"]["new 2"].key, "new 2")
        self.assertEqual(tree["new 1"]["new 2"]["__<-__"].key, "new 1")

    def test_nonodes_are_falsey(self):
        tree = ValueNode("root")
        self.assertFalse(tree["unexistent 1"]["unexistent 2"])
    
    def test_nodes_are_truthy(self):
        tree = ValueNode("root")
        tree.new_leaf("existent 1")
        self.assertTrue(tree["existent 1"])


if __name__ == "__main__":
    unittest.main()
